@@ -2,7 +2,7 @@
 
   A termios library for Ruby.
   Copyright (C) 1999, 2000, 2002 akira yamada.
-  $Id: termios.c,v 1.8 2002/10/12 15:15:25 akira Exp $
+  $Id: termios.c,v 1.10 2007-05-30 04:09:04 akira Exp $
 
  */
 
@@ -12,6 +12,18 @@
 #include <unistd.h>
 #include <string.h>
 
+#ifdef HAVE_TYPE_RB_IO_T
+typedef rb_io_t OpenFile;
+#endif
+
+#if defined(HAVE_TYPE_RB_IO_T) ||  defined(HAVE_ST_FD)
+# define FILENO(fptr) (fptr->fd)
+#else
+# define FILENO(fptr) fileno(fptr->f)
+#endif
+
+#define validate_ulong(v) ULONG2NUM(NUM2ULONG(v))
+
 static VALUE mTermios;
 static VALUE cTermios;
 static VALUE tcsetattr_opt, tcflush_qs, tcflow_act;
@@ -21,8 +33,7 @@ static VALUE
 termios_set_iflag(self, value)
     VALUE self, value;
 {
-    Check_Type(value, T_FIXNUM);
-    rb_ivar_set(self, id_iflag, value);
+    rb_ivar_set(self, id_iflag, validate_ulong(value));
 
     return value;
 }
@@ -31,8 +42,7 @@ static VALUE
 termios_set_oflag(self, value)
     VALUE self, value;
 {
-    Check_Type(value, T_FIXNUM);
-    rb_ivar_set(self, id_oflag, value);
+    rb_ivar_set(self, id_oflag, validate_ulong(value));
 
     return value;
 }
@@ -41,8 +51,7 @@ static VALUE
 termios_set_cflag(self, value)
     VALUE self, value;
 {
-    Check_Type(value, T_FIXNUM);
-    rb_ivar_set(self, id_cflag, value);
+    rb_ivar_set(self, id_cflag, validate_ulong(value));
 
     return value;
 }
@@ -51,8 +60,7 @@ static VALUE
 termios_set_lflag(self, value)
     VALUE self, value;
 {
-    Check_Type(value, T_FIXNUM);
-    rb_ivar_set(self, id_lflag, value);
+    rb_ivar_set(self, id_lflag, validate_ulong(value));
 
     return value;
 }
@@ -71,8 +79,7 @@ static VALUE
 termios_set_ispeed(self, value)
     VALUE self, value;
 {
-    Check_Type(value, T_FIXNUM);
-    rb_ivar_set(self, id_ispeed, value);
+    rb_ivar_set(self, id_ispeed, validate_ulong(value));
 
     return value;
 }
@@ -81,8 +88,7 @@ static VALUE
 termios_set_ospeed(self, value)
     VALUE self, value;
 {
-    Check_Type(value, T_FIXNUM);
-    rb_ivar_set(self, id_ospeed, value);
+    rb_ivar_set(self, id_ospeed, validate_ulong(value));
 
     return value;
 }
@@ -102,13 +108,13 @@ termios_initialize(argc, argv, self)
 	rb_ary_store(cc_ary, i, INT2FIX(0));
     }
 
-    rb_ivar_set(self, id_iflag,  Qnil);
-    rb_ivar_set(self, id_oflag,  Qnil);
-    rb_ivar_set(self, id_cflag,  Qnil);
-    rb_ivar_set(self, id_lflag,  Qnil);
+    rb_ivar_set(self, id_iflag,  INT2FIX(0));
+    rb_ivar_set(self, id_oflag,  INT2FIX(0));
+    rb_ivar_set(self, id_cflag,  INT2FIX(0));
+    rb_ivar_set(self, id_lflag,  INT2FIX(0));
     rb_ivar_set(self, id_cc,     cc_ary);
-    rb_ivar_set(self, id_ispeed, Qnil);
-    rb_ivar_set(self, id_ospeed, Qnil);
+    rb_ivar_set(self, id_ispeed, INT2FIX(0));
+    rb_ivar_set(self, id_ospeed, INT2FIX(0));
 
     rb_scan_args(argc, argv, "07", 
 		 &c_iflag, &c_oflag, &c_cflag, &c_lflag, 
@@ -147,19 +153,19 @@ termios_to_Termios(t)
 
     obj = rb_funcall(cTermios, rb_intern("new"), 0);
 
-    termios_set_iflag(obj, INT2FIX(t->c_iflag));
-    termios_set_oflag(obj, INT2FIX(t->c_oflag));
-    termios_set_cflag(obj, INT2FIX(t->c_cflag));
-    termios_set_lflag(obj, INT2FIX(t->c_lflag));
+    termios_set_iflag(obj, ULONG2NUM(t->c_iflag));
+    termios_set_oflag(obj, ULONG2NUM(t->c_oflag));
+    termios_set_cflag(obj, ULONG2NUM(t->c_cflag));
+    termios_set_lflag(obj, ULONG2NUM(t->c_lflag));
 
     cc_ary = rb_ary_new2(NCCS);
     for (i = 0; i < NCCS; i++) {
-	rb_ary_store(cc_ary, i, INT2FIX(t->c_cc[i]));
+	rb_ary_store(cc_ary, i, CHR2FIX(t->c_cc[i]));
     }
     termios_set_cc(obj, cc_ary);
 
-    termios_set_ispeed(obj, INT2FIX(cfgetispeed(t)));
-    termios_set_ospeed(obj, INT2FIX(cfgetospeed(t)));
+    termios_set_ispeed(obj, ULONG2NUM(cfgetispeed(t)));
+    termios_set_ospeed(obj, ULONG2NUM(cfgetospeed(t)));
 
     return obj;
 }
@@ -172,23 +178,19 @@ Termios_to_termios(obj, t)
     int i;
     VALUE cc_ary;
 
-    t->c_iflag = FIX2INT(rb_ivar_get(obj, id_iflag));
-    t->c_oflag = FIX2INT(rb_ivar_get(obj, id_oflag));
-    t->c_cflag = FIX2INT(rb_ivar_get(obj, id_cflag));
-    t->c_lflag = FIX2INT(rb_ivar_get(obj, id_lflag));
+    t->c_iflag = NUM2ULONG(rb_ivar_get(obj, id_iflag));
+    t->c_oflag = NUM2ULONG(rb_ivar_get(obj, id_oflag));
+    t->c_cflag = NUM2ULONG(rb_ivar_get(obj, id_cflag));
+    t->c_lflag = NUM2ULONG(rb_ivar_get(obj, id_lflag));
 
     cc_ary = rb_ivar_get(obj, id_cc);
     for (i = 0; i < NCCS; i++) {
-	if (TYPE(RARRAY(cc_ary)->ptr[i]) == T_FIXNUM) {
-	    t->c_cc[i] = NUM2INT(RARRAY(cc_ary)->ptr[i]);
-	}
-	else {
-	    t->c_cc[i] = 0;
-	}
+        VALUE elt = rb_ary_entry(cc_ary, i);
+        t->c_cc[i] = NUM2CHR(elt);
     }
 
-    cfsetispeed(t, FIX2INT(rb_ivar_get(obj, id_ispeed)));
-    cfsetospeed(t, FIX2INT(rb_ivar_get(obj, id_ospeed)));
+    cfsetispeed(t, NUM2ULONG(rb_ivar_get(obj, id_ispeed)));
+    cfsetospeed(t, NUM2ULONG(rb_ivar_get(obj, id_ospeed)));
 }
 
 
@@ -201,9 +203,8 @@ termios_tcgetattr(io)
 
     Check_Type(io, T_FILE);
     GetOpenFile(io, fptr);
-    if (tcgetattr(fileno(fptr->f), &t) < 0) {
-	rb_raise(rb_eRuntimeError, 
-		 "can't get terminal parameters (%s)", strerror(errno));
+    if (tcgetattr(FILENO(fptr), &t) < 0) {
+        rb_sys_fail("tcgetattr");
     }
 
     return termios_to_Termios(&t);
@@ -229,7 +230,7 @@ termios_tcsetattr(io, opt, param)
     Check_Type(opt, T_FIXNUM);
     if (CLASS_OF(param) != cTermios) {
 	char *type = rb_class2name(CLASS_OF(param));
-	rb_raise(rb_eArgError, 
+	rb_raise(rb_eTypeError, 
 		 "wrong argument type %s (expected Termios::Termios)", 
 		 type);
     }
@@ -243,9 +244,8 @@ termios_tcsetattr(io, opt, param)
     old = termios_tcgetattr(io);
     GetOpenFile(io, fptr);
     Termios_to_termios(param, &t);
-    if (tcsetattr(fileno(fptr->f), tcsetattr_option, &t) < 0) {
-	rb_raise(rb_eRuntimeError,
-		 "can't set terminal parameters (%s)", strerror(errno));
+    if (tcsetattr(FILENO(fptr), tcsetattr_option, &t) < 0) {
+        rb_sys_fail("tcsetattr");
     }
 
     return old;
@@ -268,9 +268,8 @@ termios_tcsendbreak(io, duration)
     Check_Type(duration, T_FIXNUM);
 
     GetOpenFile(io, fptr);
-    if (tcsendbreak(fileno(fptr->f), FIX2INT(duration)) < 0) {
-	rb_raise(rb_eRuntimeError, 
-		 "can't transmits break (%s)", strerror(errno));
+    if (tcsendbreak(FILENO(fptr), FIX2INT(duration)) < 0) {
+        rb_sys_fail("tcsendbreak");
     }
 
     return Qtrue;
@@ -292,8 +291,8 @@ termios_tcdrain(io)
     Check_Type(io, T_FILE);
 
     GetOpenFile(io, fptr);
-    if (tcdrain(fileno(fptr->f)) < 0) {
-	rb_raise(rb_eRuntimeError, "can't drain (%s)", strerror(errno));
+    if (tcdrain(FILENO(fptr)) < 0) {
+        rb_sys_fail("tcdrain");
     }
 
     return Qtrue;
@@ -317,13 +316,13 @@ termios_tcflush(io, qs)
     Check_Type(qs, T_FIXNUM);
     queue_selector = FIX2INT(qs);
     if (rb_ary_includes(tcflush_qs, qs) != Qtrue) {
-	rb_raise(rb_eTypeError, 
+	rb_raise(rb_eArgError, 
 		 "wrong queue-selector value %d", queue_selector);
     }
 
     GetOpenFile(io, fptr);
-    if (tcflush(fileno(fptr->f), queue_selector) < 0) {
-	rb_raise(rb_eRuntimeError, "can't flush (%s)", strerror(errno));
+    if (tcflush(FILENO(fptr), queue_selector) < 0) {
+        rb_sys_fail("tcflush");
     }
 
     return Qtrue;
@@ -352,9 +351,8 @@ termios_tcflow(io, act)
     }
 
     GetOpenFile(io, fptr);
-    if (tcflow(fileno(fptr->f), action) < 0) {
-	rb_raise(rb_eRuntimeError, 
-		 "can't control transmitting data flow (%s)", strerror(errno));
+    if (tcflow(FILENO(fptr), action) < 0) {
+        rb_sys_fail("tcflow");
     }
 
     return Qtrue;
@@ -372,16 +370,15 @@ termios_tcgetpgrp(io)
     VALUE io;
 {
     OpenFile *fptr;
-    int pid;
+    pid_t pid;
 
     Check_Type(io,  T_FILE);
     GetOpenFile(io, fptr);
-    if ((pid = tcgetpgrp(fileno(fptr->f))) < 0) {
-	rb_raise(rb_eRuntimeError, 
-		 "can't get process group id (%s)", strerror(errno));
+    if ((pid = tcgetpgrp(FILENO(fptr))) < 0) {
+        rb_sys_fail("tcgetpgrp");
     }
 
-    return INT2FIX(pid);
+    return LONG2NUM(pid);
 }
 
 static VALUE
@@ -396,14 +393,14 @@ termios_tcsetpgrp(io, pgrpid)
     VALUE io, pgrpid;
 {
     OpenFile *fptr;
+    pid_t pgrp;
 
     Check_Type(io,     T_FILE);
-    Check_Type(pgrpid, T_FIXNUM);
+    pgrp = NUM2LONG(pgrpid);
 
     GetOpenFile(io, fptr);
-    if (tcsetpgrp(fileno(fptr->f), FIX2INT(pgrpid)) < 0) {
-	rb_raise(rb_eRuntimeError, 
-		 "can't set process group id (%s)", strerror(errno));
+    if (tcsetpgrp(FILENO(fptr), pgrp) < 0) {
+        rb_sys_fail("tcsetpgrp");
     }
 
     return Qtrue;
@@ -425,46 +422,65 @@ termios_s_newtermios(argc, argv, klass)
     return rb_funcall2(cTermios, rb_intern("new"), argc, argv);
 }
 
+static VALUE
+termios_dup(self)
+    VALUE self;
+{
+    VALUE result;
+    VALUE cc_ary;
+    
+    result = rb_call_super(0, 0);
+    cc_ary = rb_ivar_get(self, id_cc);
+    rb_ivar_set(result, id_cc, rb_ary_dup(cc_ary));
+
+    return result;
+}
+
 void
 Init_termios()
 {
-    VALUE ccindex, iflags, oflags, cflags, lflags, bauds;
+    VALUE ccindex, ccindex_names;
+    VALUE iflags, iflags_names;
+    VALUE oflags, oflags_names, oflags_choices;
+    VALUE cflags, cflags_names, cflags_choices;
+    VALUE lflags, lflags_names;
+    VALUE bauds, bauds_names;
 
     /* module Termios */
 
     mTermios = rb_define_module("Termios");
 
-    rb_define_module_function(mTermios,"tcgetattr",  termios_s_tcgetattr,  1);
-    rb_define_module_function(mTermios,  "getattr",  termios_s_tcgetattr,  1);
-    rb_define_method(mTermios,         "tcgetattr",  termios_tcgetattr,    0);
+    rb_define_singleton_method(mTermios,"tcgetattr",  termios_s_tcgetattr,  1);
+    rb_define_module_function(mTermios,   "getattr",  termios_s_tcgetattr,  1);
+    rb_define_method(mTermios,          "tcgetattr",  termios_tcgetattr,    0);
 
-    rb_define_module_function(mTermios,"tcsetattr",  termios_s_tcsetattr,  3);
-    rb_define_module_function(mTermios,  "setattr",  termios_s_tcsetattr,  3);
-    rb_define_method(mTermios,         "tcsetattr",  termios_tcsetattr,    2);
+    rb_define_singleton_method(mTermios,"tcsetattr",  termios_s_tcsetattr,  3);
+    rb_define_module_function(mTermios,   "setattr",  termios_s_tcsetattr,  3);
+    rb_define_method(mTermios,          "tcsetattr",  termios_tcsetattr,    2);
 
-    rb_define_module_function(mTermios,"tcsendbreak",termios_s_tcsendbreak,2);
-    rb_define_module_function(mTermios,  "sendbreak",termios_s_tcsendbreak,2);
-    rb_define_method(mTermios,         "tcsendbreak",termios_tcsendbreak,  1);
+    rb_define_singleton_method(mTermios,"tcsendbreak",termios_s_tcsendbreak,2);
+    rb_define_module_function(mTermios,   "sendbreak",termios_s_tcsendbreak,2);
+    rb_define_method(mTermios,          "tcsendbreak",termios_tcsendbreak,  1);
 
-    rb_define_module_function(mTermios,"tcdrain",    termios_s_tcdrain,    1);
-    rb_define_module_function(mTermios,  "drain",    termios_s_tcdrain,    1);
-    rb_define_method(mTermios,         "tcdrain",    termios_tcdrain,    0);
+    rb_define_singleton_method(mTermios,"tcdrain",    termios_s_tcdrain,    1);
+    rb_define_module_function(mTermios,   "drain",    termios_s_tcdrain,    1);
+    rb_define_method(mTermios,          "tcdrain",    termios_tcdrain,      0);
 
-    rb_define_module_function(mTermios,"tcflush",    termios_s_tcflush,    2);
-    rb_define_module_function(mTermios,  "flush",    termios_s_tcflush,    2);
-    rb_define_method(mTermios,         "tcflush",    termios_tcflush,    1);
+    rb_define_singleton_method(mTermios,"tcflush",    termios_s_tcflush,    2);
+    rb_define_module_function(mTermios,   "flush",    termios_s_tcflush,    2);
+    rb_define_method(mTermios,          "tcflush",    termios_tcflush,      1);
 
-    rb_define_module_function(mTermios,"tcflow",     termios_s_tcflow,     2);
-    rb_define_module_function(mTermios,  "flow",     termios_s_tcflow,     2);
-    rb_define_method(mTermios,         "tcflow",     termios_tcflow,     1);
+    rb_define_singleton_method(mTermios,"tcflow",     termios_s_tcflow,     2);
+    rb_define_module_function(mTermios,   "flow",     termios_s_tcflow,     2);
+    rb_define_method(mTermios,          "tcflow",     termios_tcflow,       1);
 
-    rb_define_module_function(mTermios,"tcgetpgrp",  termios_s_tcgetpgrp,  1);
-    rb_define_module_function(mTermios,  "getpgrp",  termios_s_tcgetpgrp,  1);
-    rb_define_method(mTermios,         "tcgetpgrp",  termios_tcgetpgrp,  0);
+    rb_define_singleton_method(mTermios,"tcgetpgrp",  termios_s_tcgetpgrp,  1);
+    rb_define_module_function(mTermios,   "getpgrp",  termios_s_tcgetpgrp,  1);
+    rb_define_method(mTermios,          "tcgetpgrp",  termios_tcgetpgrp,    0);
 
-    rb_define_module_function(mTermios,"tcsetpgrp",  termios_s_tcsetpgrp,  2);
-    rb_define_module_function(mTermios,  "setpgrp",  termios_s_tcsetpgrp,  2);
-    rb_define_method(mTermios,         "tcsetpgrp",  termios_tcsetpgrp,  1);
+    rb_define_singleton_method(mTermios,"tcsetpgrp",  termios_s_tcsetpgrp,  2);
+    rb_define_module_function(mTermios,   "setpgrp",  termios_s_tcsetpgrp,  2);
+    rb_define_method(mTermios,          "tcsetpgrp",  termios_tcsetpgrp,    1);
 
     rb_define_module_function(mTermios,"new_termios",termios_s_newtermios, -1);
 
@@ -480,15 +496,17 @@ Init_termios()
     id_ispeed = rb_intern("@ispeed");
     id_ospeed = rb_intern("@ospeed");
 
-    rb_attr(cTermios, rb_intern("iflag"),  1, 0, Qtrue);
-    rb_attr(cTermios, rb_intern("oflag"),  1, 0, Qtrue);
-    rb_attr(cTermios, rb_intern("cflag"),  1, 0, Qtrue);
-    rb_attr(cTermios, rb_intern("lflag"),  1, 0, Qtrue);
-    rb_attr(cTermios, rb_intern("cc"),     1, 0, Qtrue);
-    rb_attr(cTermios, rb_intern("ispeed"), 1, 0, Qtrue);
-    rb_attr(cTermios, rb_intern("ospeed"), 1, 0, Qtrue);
+    rb_attr(cTermios, rb_intern("iflag"),  1, 0, Qfalse);
+    rb_attr(cTermios, rb_intern("oflag"),  1, 0, Qfalse);
+    rb_attr(cTermios, rb_intern("cflag"),  1, 0, Qfalse);
+    rb_attr(cTermios, rb_intern("lflag"),  1, 0, Qfalse);
+    rb_attr(cTermios, rb_intern("cc"),     1, 0, Qfalse);
+    rb_attr(cTermios, rb_intern("ispeed"), 1, 0, Qfalse);
+    rb_attr(cTermios, rb_intern("ospeed"), 1, 0, Qfalse);
 
     rb_define_private_method(cTermios, "initialize", termios_initialize, -1);
+    rb_define_method(cTermios, "dup", termios_dup, 0);
+    rb_define_method(cTermios, "clone", termios_dup, 0);
 
     rb_define_method(cTermios, "iflag=",  termios_set_iflag,  1);
     rb_define_method(cTermios, "oflag=",  termios_set_oflag,  1);
@@ -516,24 +534,41 @@ Init_termios()
     /* constants under Termios module */
 
     rb_define_const(mTermios, "NCCS",    INT2FIX(NCCS));
+    rb_define_const(mTermios, "POSIX_VDISABLE", INT2FIX(_POSIX_VDISABLE));
 
     ccindex = rb_hash_new();
+    ccindex_names = rb_ary_new();
     rb_define_const(mTermios, "CCINDEX", ccindex);
+    rb_define_const(mTermios, "CCINDEX_NAMES", ccindex_names);
 
     iflags = rb_hash_new();
+    iflags_names = rb_ary_new();
     rb_define_const(mTermios, "IFLAGS", iflags);
+    rb_define_const(mTermios, "IFLAG_NAMES", iflags_names);
 
     oflags = rb_hash_new();
+    oflags_names = rb_ary_new();
+    oflags_choices = rb_hash_new();
     rb_define_const(mTermios, "OFLAGS", oflags);
+    rb_define_const(mTermios, "OFLAG_NAMES", oflags_names);
+    rb_define_const(mTermios, "OFLAG_CHOICES", oflags_choices);
 
     cflags = rb_hash_new();
+    cflags_names = rb_ary_new();
+    cflags_choices = rb_hash_new();
     rb_define_const(mTermios, "CFLAGS", cflags);
+    rb_define_const(mTermios, "CFLAG_NAMES", cflags_names);
+    rb_define_const(mTermios, "CFLAG_CHOICES", cflags_choices);
 
     lflags = rb_hash_new();
+    lflags_names = rb_ary_new();
     rb_define_const(mTermios, "LFLAGS", lflags);
+    rb_define_const(mTermios, "LFLAG_NAMES", lflags_names);
 
     bauds = rb_hash_new();
+    bauds_names = rb_ary_new();
     rb_define_const(mTermios, "BAUDS", bauds);
+    rb_define_const(mTermios, "BAUD_NAMES", bauds_names);
 
     tcsetattr_opt = rb_ary_new();
     rb_define_const(mTermios, "SETATTR_OPTS", tcsetattr_opt);
@@ -546,15 +581,30 @@ Init_termios()
 
 #define define_flag(hash, flag) \
     { \
-      rb_define_const(mTermios, #flag, INT2FIX(flag)); \
+      rb_define_const(mTermios, #flag, ULONG2NUM(flag)); \
       rb_hash_aset(hash, rb_const_get(mTermios, rb_intern(#flag)), \
 	  ID2SYM(rb_intern(#flag))); \
+      rb_ary_push(hash##_names, ID2SYM(rb_intern(#flag)));\
     }
 #define define_flag2(ary, flag) \
     { \
       rb_define_const(mTermios, #flag, INT2FIX(flag)); \
       rb_ary_push(ary, rb_const_get(mTermios, rb_intern(#flag)));\
     }
+#define define_choice(hash, mask, value) \
+    { \
+      VALUE a; \
+      rb_define_const(mTermios, #value, ULONG2NUM(value)); \
+      rb_hash_aset(hash, rb_const_get(mTermios, rb_intern(#value)), \
+	  ID2SYM(rb_intern(#value))); \
+      rb_ary_push(hash##_names, ID2SYM(rb_intern(#value)));\
+      a = rb_hash_aref(hash##_choices, ID2SYM(rb_intern(#mask))); \
+      if (a == Qnil) { \
+          a = rb_ary_new(); \
+          rb_hash_aset(hash##_choices, ID2SYM(rb_intern(#mask)), a); \
+      } \
+      rb_ary_push(a, ID2SYM(rb_intern(#value))); \
+    }
 
     /* c_cc characters */
 #ifdef VINTR
@@ -572,11 +622,11 @@ Init_termios()
 #ifdef VEOF
     define_flag(ccindex, VEOF);
 #endif
-#ifdef VTIME
-    define_flag(ccindex, VTIME);
+#ifdef VEOL
+    define_flag(ccindex, VEOL);
 #endif
-#ifdef VMIN
-    define_flag(ccindex, VMIN);
+#ifdef VEOL2
+    define_flag(ccindex, VEOL2);
 #endif
 #ifdef VSWTC
     define_flag(ccindex, VSWTC);
@@ -590,8 +640,8 @@ Init_termios()
 #ifdef VSUSP
     define_flag(ccindex, VSUSP);
 #endif
-#ifdef VEOL
-    define_flag(ccindex, VEOL);
+#ifdef VDSUSP
+    define_flag(ccindex, VDSUSP);
 #endif
 #ifdef VREPRINT
     define_flag(ccindex, VREPRINT);
@@ -605,8 +655,14 @@ Init_termios()
 #ifdef VLNEXT
     define_flag(ccindex, VLNEXT);
 #endif
-#ifdef VEOL2
-    define_flag(ccindex, VEOL2);
+#ifdef VSTATUS
+    define_flag(ccindex, VSTATUS);
+#endif
+#ifdef VTIME
+    define_flag(ccindex, VTIME);
+#endif
+#ifdef VMIN
+    define_flag(ccindex, VMIN);
 #endif
 
     /* c_iflag bits */
@@ -637,21 +693,24 @@ Init_termios()
 #ifdef ICRNL
     define_flag(iflags, ICRNL);
 #endif
-#ifdef IUCLC
-    define_flag(iflags, IUCLC);
-#endif
 #ifdef IXON
     define_flag(iflags, IXON);
 #endif
-#ifdef IXANY
-    define_flag(iflags, IXANY);
-#endif
 #ifdef IXOFF
     define_flag(iflags, IXOFF);
 #endif
+#ifdef IUCLC
+    define_flag(iflags, IUCLC);
+#endif
+#ifdef IXANY
+    define_flag(iflags, IXANY);
+#endif
 #ifdef IMAXBEL
     define_flag(iflags, IMAXBEL);
 #endif
+#ifdef IUTF8
+    define_flag(iflags, IUTF8);
+#endif
 
     /* c_oflag bits */
 #ifdef OPOST
@@ -660,12 +719,12 @@ Init_termios()
 #ifdef OLCUC
     define_flag(oflags, OLCUC);
 #endif
-#ifdef ONLCR
-    define_flag(oflags, ONLCR);
-#endif
 #ifdef OCRNL
     define_flag(oflags, OCRNL);
 #endif
+#ifdef ONLCR
+    define_flag(oflags, ONLCR);
+#endif
 #ifdef ONOCR
     define_flag(oflags, ONOCR);
 #endif
@@ -678,74 +737,80 @@ Init_termios()
 #ifdef OFDEL
     define_flag(oflags, OFDEL);
 #endif
+#ifdef ONOEOT
+    define_flag(oflags, ONOEOT);
+#endif
+#ifdef OXTABS
+    define_flag(oflags, OXTABS);
+#endif
 #ifdef NLDLY
     define_flag(oflags, NLDLY);
 #endif
 #ifdef NL0
-    define_flag(oflags, NL0);
+    define_choice(oflags, NLDLY, NL0);
 #endif
 #ifdef NL1
-    define_flag(oflags, NL1);
+    define_choice(oflags, NLDLY, NL1);
 #endif
 #ifdef CRDLY
     define_flag(oflags, CRDLY);
 #endif
 #ifdef CR0
-    define_flag(oflags, CR0);
+    define_choice(oflags, CRDLY, CR0);
 #endif
 #ifdef CR1
-    define_flag(oflags, CR1);
+    define_choice(oflags, CRDLY, CR1);
 #endif
 #ifdef CR2
-    define_flag(oflags, CR2);
+    define_choice(oflags, CRDLY, CR2);
 #endif
 #ifdef CR3
-    define_flag(oflags, CR3);
+    define_choice(oflags, CRDLY, CR3);
 #endif
 #ifdef TABDLY
     define_flag(oflags, TABDLY);
 #endif
 #ifdef TAB0
-    define_flag(oflags, TAB0);
+    define_choice(oflags, TABDLY, TAB0);
 #endif
 #ifdef TAB1
-    define_flag(oflags, TAB1);
+    define_choice(oflags, TABDLY, TAB1);
 #endif
 #ifdef TAB2
-    define_flag(oflags, TAB2);
+    define_choice(oflags, TABDLY, TAB2);
 #endif
 #ifdef TAB3
-    define_flag(oflags, TAB3);
+    define_choice(oflags, TABDLY, TAB3);
 #endif
 #ifdef XTABS
-    define_flag(oflags, XTABS);
+    define_choice(oflags, TABDLY, XTABS);
 #endif
 #ifdef BSDLY
     define_flag(oflags, BSDLY);
 #endif
 #ifdef BS0
-    define_flag(oflags, BS0);
+    define_choice(oflags, BSDLY, BS0);
 #endif
 #ifdef BS1
-    define_flag(oflags, BS1);
+    define_choice(oflags, BSDLY, BS1);
 #endif
 #ifdef VTDLY
     define_flag(oflags, VTDLY);
 #endif
 #ifdef VT0
-    define_flag(oflags, VT0);
+    define_choice(oflags, VTDLY, VT0);
 #endif
 #ifdef VT1
-    define_flag(oflags, VT1);
+    define_choice(oflags, VTDLY, VT1);
 #endif
 #ifdef FFDLY
     define_flag(oflags, FFDLY);
 #endif
 #ifdef FF0
-    define_flag(oflags, FF0);
+    define_choice(oflags, FFDLY, FF0);
 #endif
 #ifdef FF1
-    define_flag(oflags, FF1);
+    define_choice(oflags, FFDLY, FF1);
 #endif
 
     /* c_cflag bit meaning */
@@ -806,36 +871,36 @@ Init_termios()
 #ifdef EXTB
     define_flag(cflags, EXTB);
 #endif
-#ifdef CSIZE
-    define_flag(cflags, CSIZE);
-#endif
-#ifdef CS5
-    define_flag(cflags, CS5);
-#endif
-#ifdef CS6
-    define_flag(cflags, CS6);
-#endif
-#ifdef CS7
-    define_flag(cflags, CS7);
-#endif
-#ifdef CS8
-    define_flag(cflags, CS8);
-#endif
-#ifdef CSTOPB
-    define_flag(cflags, CSTOPB);
-#endif
-#ifdef CREAD
-    define_flag(cflags, CREAD);
-#endif
 #ifdef PARENB
     define_flag(cflags, PARENB);
 #endif
 #ifdef PARODD
     define_flag(cflags, PARODD);
 #endif
+#ifdef CSIZE
+    define_flag(cflags, CSIZE);
+#endif
+#ifdef CS5
+    define_choice(cflags, CSIZE, CS5);
+#endif
+#ifdef CS6
+    define_choice(cflags, CSIZE, CS6);
+#endif
+#ifdef CS7
+    define_choice(cflags, CSIZE, CS7);
+#endif
+#ifdef CS8
+    define_choice(cflags, CSIZE, CS8);
+#endif
 #ifdef HUPCL
     define_flag(cflags, HUPCL);
 #endif
+#ifdef CSTOPB
+    define_flag(cflags, CSTOPB);
+#endif
+#ifdef CREAD
+    define_flag(cflags, CREAD);
+#endif
 #ifdef CLOCAL
     define_flag(cflags, CLOCAL);
 #endif
@@ -854,12 +919,48 @@ Init_termios()
 #ifdef B460800
     define_flag(bauds, B460800);
 #endif
+#ifdef B500000
+    define_flag(bauds, B500000);
+#endif
+#ifdef B576000
+    define_flag(bauds, B576000);
+#endif
+#ifdef B921600
+    define_flag(bauds, B921600);
+#endif
+#ifdef B1000000
+    define_flag(bauds, B1000000);
+#endif
+#ifdef B1152000
+    define_flag(bauds, B1152000);
+#endif
+#ifdef B1500000
+    define_flag(bauds, B1500000);
+#endif
+#ifdef B2000000
+    define_flag(bauds, B2000000);
+#endif
+#ifdef B2500000
+    define_flag(bauds, B2500000);
+#endif
+#ifdef B3000000
+    define_flag(bauds, B3000000);
+#endif
+#ifdef B3500000
+    define_flag(bauds, B3500000);
+#endif
+#ifdef B4000000
+    define_flag(bauds, B4000000);
+#endif
 #ifdef CIBAUD
     define_flag(cflags, CIBAUD);
 #endif
 #ifdef CRTSCTS
     define_flag(cflags, CRTSCTS);
 #endif
+#ifdef MDMBUF
+    define_flag(cflags, MDMBUF);
+#endif
 
     /* c_lflag bits */
 #ifdef ISIG
@@ -868,8 +969,8 @@ Init_termios()
 #ifdef ICANON
     define_flag(lflags, ICANON);
 #endif
-#ifdef XCASE
-    define_flag(lflags, XCASE);
+#ifdef IEXTEN
+    define_flag(lflags, IEXTEN);
 #endif
 #ifdef ECHO
     define_flag(lflags, ECHO);
@@ -886,15 +987,18 @@ Init_termios()
 #ifdef NOFLSH
     define_flag(lflags, NOFLSH);
 #endif
+#ifdef XCASE
+    define_flag(lflags, XCASE);
+#endif
 #ifdef TOSTOP
     define_flag(lflags, TOSTOP);
 #endif
-#ifdef ECHOCTL
-    define_flag(lflags, ECHOCTL);
-#endif
 #ifdef ECHOPRT
     define_flag(lflags, ECHOPRT);
 #endif
+#ifdef ECHOCTL
+    define_flag(lflags, ECHOCTL);
+#endif
 #ifdef ECHOKE
     define_flag(lflags, ECHOKE);
 #endif
@@ -904,8 +1008,14 @@ Init_termios()
 #ifdef PENDIN
     define_flag(lflags, PENDIN);
 #endif
-#ifdef IEXTEN
-    define_flag(lflags, IEXTEN);
+#ifdef ALTWERASE
+    define_flag(lflags, ALTWERASE);
+#endif
+#ifdef EXTPROC
+    define_flag(lflags, EXTPROC);
+#endif
+#ifdef NOKERNINFO
+    define_flag(lflags, NOKERNINFO);
 #endif
 
     /* tcflow() and TCXONC use these */
